from __future__ import annotations


from ._base import Group, LieGroup
from ._su import generalized_gell_mann, SU
from ._so import SO
from ._sp import Sp
from ._cyclic import Z

__all__ = [
    "generalized_gell_mann", "SU", "SO", "Sp", "Group", "LieGroup" ,"Z"
]


# Compiled extension imports
from .liesym import (
    debug_mode as _debug_mode,
    LieAlgebraBackend as _LieAlgebraBackend,
)

from .algebras import *
from .groups import *

__all__ = [
    "A", "B", "C", "D", "F4", "G2", "E", "LieAlgebra", "NumericSymbol",
    "SU", "generalized_gell_mann", "SO", "Sp", "Group", "LieGroup", "Z"
]



from __future__ import annotations

from typing import List, Tuple, Union
from sympy import Matrix, zeros, I, KroneckerDelta

from ._base import LieGroup
from ..algebras import B, D



class SO(LieGroup):
    """The Special Orthogonal Group
    """

    def __new__(cls, dim: int):
        if dim < 2:
            raise NotImplementedError("SO(1)==O(1) is not implemented yet.")
        return super().__new__(cls, "SO", dim)

    def __init__(self, *args, **kwargs):
        n = self.dimension
        if n.is_even:
            self._algebra = D(n / 2)
        else:
            self._algebra = B((n-1) / 2)


    def generators(self, indexed=False) -> List[Union[Matrix, Tuple[Matrix, tuple]]]:
        """Generators for SO(N).

        Args:
            indexed (bool, Optional): For N > 3, there exists a naming scheme for generators. If True returns a tuple
            of the matrix and its (m,n) index.

        Returns:
            List[Union[Matrix, Tuple[Matrix, tuple]]]: List of (mathematical) generators

        Sources:
            - http://www.astro.sunysb.edu/steinkirch/books/group.pdf
        """
        results = []
        for m in range(self.dimension):
            for n in range(m):                      
                mat = zeros(self.dimension)          
                for i in range(self.dimension):
                    for j in range(self.dimension):
                        mat[i,j] = - I * (KroneckerDelta(m,i) * KroneckerDelta(n,j) - KroneckerDelta(m,j) * KroneckerDelta(n,i))
                results.append((mat, (m,n)) if indexed else mat)
        if indexed:
            return sorted(results, key=lambda y: y[1])
        return results


# from __future__ import annotations
from functools import reduce

from sympy import exp, I, pi, Symbol

from ._base import Group


class Z(Group):
    """Cyclic Group
    """
    
    def __new__(cls, dim: int):
        return super().__new__(cls, "Z", dim)

    def generators(self, indexed=False) -> list:
        """The basis for the generators in the cyclic group are 
        in the exponential imaginary basis.

        Args:
            indexed (bool, Optional): Returns tuple of named generators. Defaults to False.

        Examples
        =========
        >>> from liesym import Z
        >>> Z(3).generators()
        [1, exp(2*I*pi/3), exp(-2*I*pi/3)]
        >>> Z(3).generators(indexed=True)
        [1, exp(2*I*pi/3), exp(-2*I*pi/3)]
        """
        d = self.dimension
        
        gens = [exp(2 * pi * I * x / d) for x in range(d)]
        if indexed:
            return [(Symbol(f"Z_{idx}"), x)  for idx, x in gens]
        return gens


    def product(self, *args, **kwargs) -> list:
        """Product of the Cyclic representations

        Returns:
            list: To keep types standard, returns a list of length 1

        Examples
        =========
        >>> from liesym import Z
        >>> z5 = Z(5)
        >>> g = z5.generators()
        >>> z5.product(g[0], g[1], g[2], g[3], g[4])
        [1]
        """
        return [reduce(lambda a,b: a*b,args)]

    def sym_product(self, *args, as_tuple=False, **kwargs) -> list:
        """Will take the product symbolically of Cyclic representations
        of the form `lambda x: f`Z_{x}`

        Returns:
            list: To keep types standard, returns a list of length 1

        Examples
        =========
        >>> from liesym import Z
        >>> z5 = Z(5)
        >>> z5.sym_product("Z_1", "Z_2")
        [Z_3]
        >>> z5.sym_product("Z_3", "Z_4", as_tuple=True)
        [(Z_2, exp(4*I*pi/5))]
        """

        cleaned_args = [Symbol(x) if isinstance(x,str) else x for x in args]

        gens = {Symbol(f"Z_{idx}"): x  for idx, x in enumerate(self.generators())}
        
        result = self.product(*[gens[x] for x in cleaned_args])
        
        for k,v in gens.items():
            if v == result[0]:
                if as_tuple:
                    return [(k,v)]
                return [k]
        raise ValueError("Malformed cyclic product")



from __future__ import annotations

from typing import Iterator
from sympy import Matrix, zeros, I, sqrt

from ._base import LieGroup
from ..algebras import A


def generalized_gell_mann(dimension: int) -> Iterator[Matrix]:
    """Generator of generalized Gell-Mann Matrices.

    Args:
        dimension (int): Dimension of group

    Yields:
        Iterator[Matrix]: List of (mathematical) generators

    Sources:
        - https://mathworld.wolfram.com/GeneralizedGell-MannMatrix.html

    """
    def eij(dim):    
            def _(i,j):
                mat = zeros(dim)
                mat[i,j] = 1
                return mat
            return _
    E = eij(dimension)

    # Symmetric
    for k in range(dimension):
        for j in range(k):
            yield E(k,j) + E(j,k)

    # AntiSymmetric
    for k in range(dimension):
        for j in range(k):
            yield I * (E(k,j) - E(j,k))

    # Diagonal
    for l in range(dimension - 1):
        coeff =  sqrt(2) / sqrt((l+1) * (l + 2))
        sum_term =  sum([E(j,j) for j in range(l+1)], zeros(dimension))
        yield coeff * (sum_term - (l+1) * E(l+1, l+1))
    

class SU(LieGroup):
    """The Special Unitary Group
    """

    def __new__(cls, dim: int):
        if dim < 2:
            raise NotImplementedError("SU(1)==U(1) is not implemented yet.")
        return super().__new__(cls, "SU", dim)

    def __init__(self, *args, **kwargs):
        self._algebra = A(self.dimension - 1)

    def generators(self):
        """ Generators for SU(N) in the basis of Gell-Mann
        """
        return list(generalized_gell_mann(self.dimension))




from __future__ import annotations

from typing import List, Tuple, Union
from sympy.core import Basic
from sympy import Matrix, Symbol, sympify

from ..algebras import LieAlgebra


class Group(Basic):
    """The base class for all (lie) groups. The methods and properties
    in this class are basis independent and apply in a general sense. In
    order to write down the roots as matricies and vectors, we choose a 
    representation.
    """

    def __new__(cls, group: str, dim: int):
        obj = super().__new__(cls, sympify(dim))
        obj._group = group
        return obj

    @property
    def group(self) -> str:
        """Group type
        """
        return self._group

    @property
    def dimension(self) -> int:
        """Group dimension
        """
        return self.args[0]

    def generators(self) -> list:
        """Generalized matrix generators of the group.

        Abstract
        """
        pass

    def product(self, *args, **kwargs) -> list:
        """Calculates the product between several group representations.
        If backing algebra is a Lie Algebra, defaults to `LieAlgebra.tensor_product_decomposition`.

        Abstract
        """
        pass

    def sym_product(self, *args, as_tuple=False, **kwargs) -> list:
        """Calculates the product between several group representations passed by the symbolic name.
        If backing algebra is a Lie Algebra, defaults to `LieAlgebra.tensor_product_decomposition`

        Abstract
        """
        pass

class LieGroup(Group):
    """Group that has a Lie Algebra associated with it.
    """

    def __init__(self, *args, **kwargs):
        """Used to set lazy properties
        """
        self._algebra = None


    @property
    def algebra(self) -> LieAlgebra:
        """Backing Lie Algebra
        """
        return self._algebra

    def product(self, *args, **kwargs) -> List['Matrix']:
        """Uses tensor product decomposition to find the products between the 
        representations. Supported kwargs can be found on `LieAlgebra.tensor_product_decomposition`

        Returns:
            List[Matrix]: Tensor sum of the product.

        Examples
        =========
        >>> from liesym import SO
        >>> from sympy import Matrix
        >>> so10 = SO(10)
        >>> so10.product(Matrix([[1,0,0,0,0]]),Matrix([[1,0,0,0,0]]))
        >>> print(results)
        [Matrix([[0, 0, 0, 0, 0]]),
         Matrix([[0, 1, 0, 0, 0]]),
         Matrix([[2, 0, 0, 0, 0]])]

        """
        
        return self.algebra.tensor_product_decomposition(args, **kwargs) # type: ignore

    def sym_product(self, *args, as_tuple=False, **kwargs) -> List[Union[Symbol, Tuple[Matrix, Symbol]]]:
        """Uses tensor product decomposition to find the products between the 
        representations. Supported kwargs can be found on `LieAlgebra.tensor_product_decomposition`.

        Args:
            as_tuple (bool, optional): If True, returns tuple of matrix rep with symbol. Defaults to False.

        Returns:
            List[Union[Symbol, Tuple[Matrix, Symbol]]]: List of symbols or list of (Matrix, Symbol)

        Examples
        ========
        >>> from liesym import SO
        >>> so10 = SO(10)
        >>> so10.sym_product("10","45")        
        [120, 10, 320]
        >>> so10.sym_product("10","45", as_tuple=True)
        [(Matrix([[0, 0, 1, 0, 0]]), 120),
         (Matrix([[1, 0, 0, 0, 0]]), 10),
         (Matrix([[1, 1, 0, 0, 0]]), 320)]
        >>> from liesym import SU
        >>> su2 = SU(2)
        >>> su.sym_product('3', '\\bar{3}')
        [1, 8]
        >>> su.sym_product('3', '\\bar{3}', '3', as_tuple=True)
        [(Matrix([[1, 0]]), 3),
        (Matrix([[0, 2]]), '\\bar{6}'),
        (Matrix([[1, 0]]), 3),
        (Matrix([[2, 1]]), 15)]
        """
        mats = [self.algebra.irrep_lookup(x) for x in args]

        results = []
        for rep in self.algebra.tensor_product_decomposition(mats, **kwargs):
            symbol = self.algebra.dim_name(rep)
            if as_tuple:
                results.append((rep, symbol))
            else:
                results.append(symbol)
        return results
